/*  This is kbdlock
 *  Copyright (C) 1999,2005,2012 Sergey Poznyakoff
 *
 *  Kbdlock is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Kbdlock is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with kbdlock; if not, see <http://www.gnu.org/licenses/>.
 */
#define _GNU_SOURCE
#include <getopt.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <linux/vt.h>
#include <pwd.h>
#include <signal.h>
#include <errno.h>
#include <crypt.h>
#include <time.h>
#ifdef HAVE_CURSES
# include <curses.h>
# include <term.h>
#endif
#include <sys/utsname.h>
#include <utmp.h>
#include <termios.h>
#ifdef __SHADOW__
# include <shadow.h>
#endif

#ifndef TCSASOFT
# define TCSASOFT 0
#endif

char *copyleft = "\
Copyright (C) 1999,2005,2012 Sergey Poznyakoff\n\
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n\
This is free software: you are free to change and redistribute it.\n\
There is NO WARRANTY, to the extent permitted by law.\n\
\n\
Written by Sergey Poznyakoff.\n\
";

char *device = "/dev/tty";
char *issue = "/etc/issue";

#ifdef GNU_STYLE_OPTIONS
typedef struct option LONGOPT;
/* Note: Internal variables go first */
LONGOPT longopts[] = {
    "help",    no_argument, 0, 'h',
    "issue",   required_argument, 0, 'i',
    "clear", no_argument, 0, 'c',
    "license", no_argument, 0, 'l',
    "password", no_argument, 0, 'p',
    "sleep", required_argument, 0, 's',
    "timeout", required_argument, 0, 't',
    "version", no_argument, 0, 'v',
    "quiet", no_argument, 0, 'q',
    0,
};
#else
#define getopt_long(argc, argv, optstr, long_opts, iptr) \
 getopt(argc, argv, optstr)
#endif

char *progname;
char usage[] =
"usage: kbdlock [OPTIONS]\n"
"Linux console locker\n"
"OPTIONS are:\n"
#ifdef HAVE_CURSES
"      -c, --clear                 clear screen before the prompt\n"    
#endif
"      -i, --issue=FILE            show content of FILE instead of /etc/issue\n"
"      -p, --password              prompt for password to use\n"
"      -s, --sleep=N               sleep N seconds after each wrong password\n"
"                                  (default 3)\n"    
"      -t, --timeout=N             set timeout N seconds\n"
"      -q, --quiet                 be quiet\n"
"Informational options:\n"
"      -h, --help                  display this help screen\n"
"      -v, --version               display version\n"
"\n"
"Report bugs to: <gray@gnu.org.ua>.\n";

int quiet = 0;
int timeout_option = 0;
long time_left;
static char inputbuf[256];
long sleep_time = 3;
int clear_option;

static long gettimeout(char *str);
static int readpass(int, int);

main(argc, argv)
int argc;
char **argv;
{
    int fd, c, rc;
    struct termios to, tn;
    uid_t uid = getuid();
    struct passwd *pwd;
#ifdef __SHADOW__
    struct spwd *spwd;
#endif    
    char *p1, *p2, *username;
    char *pass = NULL;
    char salt[3];
    
    progname = argv[0];

    while ((c = getopt_long(argc, argv, "chi:ps:t:vq", longopts, &rc))
	   != EOF) {
	switch (c) {
	case 'c':
	    clear_option = 1;
	    break;
	case 'h':
	    printf("%s", usage);
	    exit(0);
	case 'i':
	    issue = optarg;
	    break;
	case 'p':
	    p1 = strdup(getpass("password:"));
	    p2 = strdup(getpass("re-enter:"));
	    if (strcmp(p1, p2)) {
		printf("entries differ");
		exit(1);
	    }
	    memset(p1, 0, strlen(p1));
	    free(p1);

	    srand(time(NULL));
	    strcpy(salt, "$1");
	    pass = strdup(crypt(p2, salt));

            memset(p2, 0, strlen(p2));
	    free(p2);
	    break;
	case 's':
	    sleep_time = gettimeout(optarg);
	    break;
	case 't':
	    time_left = gettimeout(optarg);
	    timeout_option = 1;
	    break;
	case 'v':
	    printf("kbdlock %s\n", VERSION);
	    printf("%s\n", copyleft);
	    exit(0);
	case 'q':
	    quiet++;
	    break;
	default:
	    printf("%s", usage);
	    exit(1);
	}
    }

    if (optind < argc) {
	fprintf(stderr, "%s: unexpected argument\n", progname);
	exit(1);
    }
    
    pwd = getpwuid(uid);
    username = pwd->pw_name;
    if (!pass) {
#ifdef __SHADOW__
	spwd = getspnam(username);
	if (!spwd) {
	    fprintf(stderr, "%s has wrong permissions\n", progname);
	    exit(1);
	}
	pass = strdup(spwd->sp_pwdp);
#else
	if (!pwd->pw_passwd[0] || strlen(pwd->pw_passwd) == 1) {
	    fprintf(stderr, "%s has wrong permissions\n", progname);
	    exit(1);
	}
	pass = pwd->pw_passwd;
#endif
    }
    
    setvbuf(stdout, NULL, _IONBF, 0);
	
#if !defined(__DEBUG__)
    signal(SIGHUP, SIG_IGN);
    signal(SIGINT, SIG_IGN);
    signal(SIGQUIT, SIG_IGN);
    signal(SIGILL, SIG_IGN);
    signal(SIGTRAP, SIG_IGN);
    signal(SIGABRT, SIG_IGN);
    signal(SIGBUS, SIG_IGN);
    signal(SIGFPE, SIG_IGN);
    signal(SIGKILL, SIG_IGN);
    signal(SIGUSR1, SIG_IGN);
    signal(SIGSEGV, SIG_IGN);
    signal(SIGUSR2, SIG_IGN);
    signal(SIGPIPE, SIG_IGN);
    signal(SIGALRM, SIG_IGN);
    signal(SIGTERM, SIG_IGN);
    signal(SIGCHLD, SIG_IGN);
    signal(SIGCONT, SIG_IGN);
    signal(SIGSTOP, SIG_IGN);
    signal(SIGTSTP, SIG_IGN);
    signal(SIGTTIN, SIG_IGN);
    signal(SIGTTOU, SIG_IGN);
    signal(SIGURG, SIG_IGN);
    signal(SIGXCPU, SIG_IGN);
    signal(SIGXFSZ, SIG_IGN);
    signal(SIGVTALRM, SIG_IGN);
    signal(SIGPROF, SIG_IGN);
    signal(SIGWINCH, SIG_IGN);
    signal(SIGIO, SIG_IGN);
    signal(SIGPWR, SIG_IGN);
#endif    
    fd = open(device, O_RDONLY);
    if (fd == -1) {
	perror(device);
	exit(1);
    }
    if (!isatty(fd)) {
    	fprintf(stderr, "%s: stdin not a terminal. Exiting...\n", progname);
    	exit(1);
    }
    rc = tcgetattr(fd, &to);
    if (rc) {
	perror("tcgetattr");
	exit(1);
    }
    tn = to;
    tn.c_lflag &= ~(ECHO | ISIG);
    rc = tcsetattr(fd, TCSAFLUSH | TCSASOFT, &tn);
    if (rc) {
	perror("tcgetattr");
	exit(1);
    }
#if !defined(__DEBUG__)
    rc = ioctl(fd, VT_LOCKSWITCH);
    /*    printf("ioctl: %d\n", rc);*/
    if (rc) {
	perror("VT_LOCKSWITCH");
	tcsetattr(fd, TCSAFLUSH | TCSASOFT, &to);
	exit(1);
    }
#endif
#ifdef HAVE_CURSES
    setupterm(NULL, 1, NULL);
#endif
    while (1) {
	if (!quiet) {
#ifdef HAVE_CURSES
	    if (clear_option)
		tputs(clear_screen, 1, putchar);
#endif
	    prissue();
	    printf("tty is locked by %s", username);
	    if (timeout_option)
		printf(" (timeout %ld sec)", time_left);
	    printf("; password: ");
	}

	if (readpass(fd, to.c_cc[VERASE])) {
	    printf("\n%s timed out\n", progname);
	    break;
	}
	putchar('\n');
	if (strcmp(crypt(inputbuf, pass), pass) == 0)
	    break;
	memset(inputbuf, 0, sizeof(inputbuf));
	printf("sorry\n");
	wait();
    }

    rc = tcsetattr(fd, TCSAFLUSH | TCSASOFT, &to);

#if !defined(__DEBUG__)
    ioctl(fd, VT_UNLOCKSWITCH);
    close(fd);
#endif
    return 0;
}

readpass(fd, erase)
{
    struct timeval tv, *tvp = timeout_option ? &tv : NULL;
    fd_set rd;
    int rc;
    char c;
    time_t start = time(NULL);
    int i = 0;

    for (i = 0; i < sizeof(inputbuf)-1; ) {
	FD_ZERO(&rd);
	FD_SET(fd, &rd);
	if (timeout_option) {
	    if (time_left <= 0)
		return 1;
	    tv.tv_sec = time_left;
	    tv.tv_usec = 0;
	}
	rc = select(fd + 1, &rd, NULL, NULL, tvp);
	if (rc < 0) {
	    if (errno != EINTR) { 
		perror("select");
		return 0;
	    }
	    continue;
	}
	if (timeout_option && ((time_left -= (time(NULL) - start))) <= 0)
	    return 1;
	if (FD_ISSET(fd, &rd)) {
	    rc = read(fd, &c, 1);
	    if (c == '\n')
		break;
	    else if (c == erase) {
		if (i > 0)
		    i--;
	    } else
		inputbuf[i++] = c;
	}
    }
    inputbuf[i] = 0;
    return 0;
}

wait()
{
    struct timeval tv;
    time_t start = time(NULL);
    int ttw = sleep_time;
    for (;;) {
	tv.tv_sec = ttw;
	tv.tv_usec = 0;
	if (select(0, NULL, NULL, NULL, &tv) < 0)
	    continue;
	if ((ttw -= (time(NULL) - start)) <= 0)
	    break;
    }
    if (timeout_option)
	time_left -= sleep_time;
}

unsigned
usrcnt()
{
    unsigned n = 0;
    struct utmp *p;

    for (setutent(); p = getutent();) {
	if (p->ut_type == USER_PROCESS)
	    ++n;
    }
    return n;
}

prissue()
{
    int c;
    FILE *fp = fopen(issue, "r");

    if (!fp)
	return;
    while ((c = fgetc(fp)) != EOF) {
	if (c == '\\') {
	    if ((c = fgetc(fp)) == EOF) {
		putchar('\\');
		break;
	    } else if (c == 'c') {
		if ((c = fgetc(fp)) == EOF)
		    break;
		if (c == '\n')
		    continue;
	    } else if (c == '\n') {
		continue;
	    } else {
		time_t t;
		char timebuf[80];
		unsigned n;
		char *p;
		struct utsname unb;
		int unb_init = 1;
		if (unb_init == 1)
		    unb_init = uname(&unb);

		switch (c) {
		case 'b':
		    /* Insert the baudrate of the current line. */
		    printf("38400"); /* FIXME */
		    break;
		case 'd':
		    /* Insert the current date. */
		    time(&t);
		    strftime(timebuf, sizeof(timebuf),
			     "%a %b %d  %Y", localtime(&t));
		    printf("%s", timebuf);
		    break;
		case 's':
		    /* Insert the system name, the name of the operating
		       system. */
		    if (!unb_init)
			printf("%s", unb.sysname);
		    break;
		case 'l':
		    /* Insert the name of the current tty line. */
		    p = ttyname(0);
		    if (strncmp(p, "/dev/", 5) == 0)
			p += 5;
		    printf("%s", p);
		    break;
		case 'm':
		    /* Insert the architecture identifier of the machine,
		       e.g., i686. */
		    if (!unb_init)
			printf("%s", unb.machine);
		    break;
		case 'n':
		    /* Insert the nodename of the machine, also known as the
		       hostname. */
		    if (!unb_init)
			printf("%s", unb.nodename);
		    break;
		case 'o':
		    /* Insert the domainname of the machine. */
		    if (!unb_init)
			printf("%s", unb.domainname);
		    break;
		case 'r':
		    /* Insert the release number of the kernel,
		       e.g., 2.6.11.12. */
		    if (!unb_init)
			printf("%s", unb.release);
		    break;
		case 't':
		    /* Insert the current time. */
		    time(&t);
		    strftime(timebuf, sizeof(timebuf),
			     "%H:%M:%S", localtime(&t));
		    printf("%s", timebuf);
		    break;
		case 'u':
		    /* Insert the number of current users logged in. */
		    printf("%u", usrcnt());
		    break;
		case 'U':
		    /* Insert the string "1 user" or "<n> users" where <n> is
		       the number of current users logged in. */
		    n = usrcnt();
		    if (n == 1)
			printf("1 user");
		    else
			printf("%u users", n);
		    break;
		case 'v':
		    /* Insert the version of the OS, e.g., the build-date etc.
		     */
		    if (!unb_init)
			printf("%s", unb.version);
		    break;		    
		default:
		    putchar(c);
		}
	    }
	} else
	    putchar(c);
    }
    fclose(fp);
}

long
gettimeout(str)
char *str;
{
    char *p;
    int n;
    long hours = 0, minutes = 0, seconds = 0;
    
    p = str;

    for (;;) {
	n = 0;
	while (*p && isdigit(*p)) {
	    n = 10*n + *p - '0';
	    p++;
	}
    
	switch (*p) {
	case 'h':
	case 'H':
	    if (hours) {
		fprintf(stderr, "%s: hours specified twice in \"%s\"\n",
			progname, str);
		exit(1);
	    }
	    hours = n;
	    break;
	case 'm':
	case 'M':
	    if (minutes) {
		fprintf(stderr, "%s: minutes specified twice in \"%s\"\n",
			progname, str);
		exit(1);
	    }
	    minutes = n;
	    break;
	case 's':
	case 'S':
	    if (seconds) {
		fprintf(stderr, "%s: seconds specified twice in \"%s\"\n",
			progname, str);
		exit(1);
	    }
	    seconds = n;
	    break;
	case 0:
	    if (!hours && !minutes && !seconds)
		return n;
	    if (n) {
		if (seconds) {
		    fprintf(stderr, "%s: seconds specified twice in \"%s\"\n",
			    progname, str);
		    exit(1);
		}
		seconds = n;
	    }
	    return (hours*60 + minutes)*60 + seconds;
	default:
	    fprintf(stderr,
		    "%s: time specification error, "
		    "^ marks the error point: %*.*s^%s\n",
		    progname,
		    p - str, p - str, str, p);
	    exit(1);
	}
	p++;
    }
}
